#include <cstdlib>
#include <cstring>
#include <ReWindow/types/Window.h>
#include <ReWindow/types/Cursors.h>
#include <ReWindow/Logger.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

class ReWindow::RWindow::Platform {
public:
    Window window;
    XEvent xev;
    Display* display;
    int defaultScreen;
    XVisualInfo* visual;
    XSetWindowAttributes xSetWindowAttributes;
    XWindowAttributes windowAttributes;
    Atom wmDeleteWindow;
    Atom windowTypeAtom;
    Atom windowTypeUtilityAtom;
    XSizeHints hints;
    Cursor invisible_cursor = 0;
};


using namespace ReWindow;

RWindow::RWindow() {
platform = new Platform();
extant = this;
}

IPair position;
RWindow::RWindow(const std::string& wTitle, int wWidth, int wHeight, bool wFullscreen, bool wResizable, bool wVsync)
        : title(wTitle), width(wWidth), height(wHeight), fullscreen_mode(wFullscreen), resizable(wResizable), vsync(wVsync),
          flags{false,wFullscreen,wResizable,wVsync} {
    platform = new Platform();
    extant = this;
}

void RWindow::Raise() {
    Logger::Debug(std::format("Raising window '{}'", this->title));
    // Get the position of the renderable area relative to the rest of the window.
    XGetWindowAttributes(platform->display, platform->window, &platform->windowAttributes);
    render_area_position =  { platform->windowAttributes.x, platform->windowAttributes.y };
    XRaiseWindow(platform->display, platform->window);
}

void RWindow::Lower()
{
    Logger::Debug(std::format("Lowering window '{}'", this->title));
    XLowerWindow(platform->display, platform->window);
}

void RWindow::DestroyOSWindowHandle() {
    Logger::Debug(std::format("Destroying window '{}'", this->title));
    XDestroySubwindows(platform->display, platform->window);
    Logger::Debug(std::format("Destroyed window '{}'", this->title));
    XDestroyWindow(platform->display, platform->window);

    system("xset r on"); // Re-set X-server parameter to enable key-repeat.

    open = false;
}

//void RWindow::

void RWindow::SetCursorVisible(bool cursor_enable) {
    cursor_visible = cursor_enable;
    if (platform->invisible_cursor == 0) {
        Pixmap blank_pixmap = XCreatePixmap(platform->display, platform->window, 1, 1, 1);
        XColor dummy; dummy.pixel = 0; dummy.red = 0; dummy.flags = 0;
        platform->invisible_cursor = XCreatePixmapCursor(platform->display, blank_pixmap, blank_pixmap, &dummy, &dummy, 0, 0);
        XFreePixmap(platform->display, blank_pixmap);
    }

    if (!cursor_enable)
        XDefineCursor(platform->display, platform->window, platform->invisible_cursor);

    if (cursor_enable)
        XUndefineCursor(platform->display, platform->window);
}

void RWindow::SetResizable(bool sizable) {
    XGetWindowAttributes(platform->display, platform->window, &platform->windowAttributes);


    this->resizable = sizable;
    if (!sizable)
    {
        Logger::Debug("Once you've done this you cannot make it resizable again.");
        platform->hints.flags = PMinSize | PMaxSize;
        platform->hints.min_width = platform->hints.max_width = platform->windowAttributes.width;
        platform->hints.min_height = platform->hints.max_height = platform->windowAttributes.height;
        XSetWMNormalHints(platform->display, platform->window, &platform->hints);
    }
    //this->SetFlag(WindowFlag::RESIZABLE, resizable);

}

void RWindow::SetFlag(WindowFlag flag, bool state) {
    XGetWindowAttributes(platform->display, platform->window, &platform->windowAttributes);
    flags[(int) flag] = state;
    //Once you've done this you cannot make it resizable again.
    if (flag == WindowFlag::RESIZABLE && !state) {
        Logger::Debug("Once you've done this you cannot make it resizable again.");
        platform->hints.flags = PMinSize | PMaxSize;
        platform->hints.min_width = platform->hints.max_width = platform->windowAttributes.width;
        platform->hints.min_height = platform->hints.max_height = platform->windowAttributes.height;
        XSetWMNormalHints(platform->display, platform->window, &platform->hints);
    }
    Logger::Debug(std::format("Set flag '{}' to state '{}' for window '{}'", WindowFlagToStr(flag), state, this->title));
}

void RWindow::PollEvents() {
    while(XPending(platform->display)) {
        XNextEvent(platform->display, &platform->xev);

        if (platform->xev.type == ClientMessage)
            Logger::Info(std::format("Event '{}'", "ClientMessage"));

        if (platform->xev.xclient.message_type == XInternAtom(platform->display, "WM_PROTOCOLS", False) && static_cast<Atom>(platform->xev.xclient.data.l[0]) == platform->wmDeleteWindow) {
            Close();
        }

        if (platform->xev.type == FocusIn) {
            Logger::Debug(std::format("Event'{}'", "FocusIn"));
            XAutoRepeatOff(platform->display);
            SetFlag(WindowFlag::IN_FOCUS, true);

            if (!cursor_visible)
                XDefineCursor(platform->display, platform->window, platform->invisible_cursor);

            // Get the position of the renderable area relative to the rest of the window.
            XGetWindowAttributes(platform->display, platform->window, &platform->windowAttributes);
            render_area_position =  { platform->windowAttributes.x, platform->windowAttributes.y };
            processFocusIn();
        }

        if (platform->xev.type == FocusOut) {
            Logger::Debug(std::format("Event '{}'", "FocusOut"));
            XAutoRepeatOn(platform->display);

            SetFlag(WindowFlag::IN_FOCUS, false);
            if (!cursor_visible)
                XUndefineCursor(platform->display, platform->window);

            processFocusOut();
        }

        if (platform->xev.type == KeyRelease) {
            Logger::Debug(std::format("Event '{}'", "KeyRelease"));
            auto scancode = (X11Scancode) platform->xev.xkey.keycode;
            auto key = GetKeyFromX11Scancode(scancode);
            processKeyRelease(key);
        }

        if (platform->xev.type == KeyPress) {
            Logger::Debug(std::format("Event '{}'", "KeyPress"));
            auto scancode = (X11Scancode) platform->xev.xkey.keycode;
            auto key = GetKeyFromX11Scancode(scancode);
            processKeyPress(key);
        }

        if (platform->xev.type == ButtonRelease) {

            // Mouse Wheel fires both the ButtonPress and ButtonRelease instantaneously.
            // Therefore, we handle it as a specific MouseWheel event rather than a MouseButton event,
            // and only call on ButtonPress, otherwise it will appear to duplicate the mouse wheel scroll.
            if (platform->xev.xbutton.button != 4 && platform->xev.xbutton.button != 5) {
                MouseButton button = GetMouseButtonFromXButton(platform->xev.xbutton.button);
                Logger::Debug(std::format("Event '{}'", "ButtonRelease"));
                processMouseRelease(button);
            }
        }

        if (platform->xev.type == ButtonPress) {

            // Mouse Wheel fires both the ButtonPress and ButtonRelease instantaneously.
            // Therefore, we handle it as a specific MouseWheel event rather than a MouseButton event,
            // and only call on ButtonPress, otherwise it will appear to duplicate the mouse wheel scroll.
            if (platform->xev.xbutton.button == 4) {
                processMouseWheel(-1);
            } else if (platform->xev.xbutton.button == 5) {
                processMouseWheel(1);
            } else
            {
                MouseButton button = GetMouseButtonFromXButton(platform->xev.xbutton.button);

                Logger::Debug(std::format("Event: MouseButtonPress {}", button.Mnemonic));

                processMousePress(button);
            }
        }

        if (platform->xev.type == Expose)
        {
            Logger::Debug(std::format("Event '{}'", "Expose"));
        }

        // NOTE: This event is functionally useless, as it only informs of the very beginning and end of a mouse movement.
        if (platform->xev.type == MotionNotify)
        {
            Logger::Debug(std::format("Event '{}'", "MotionNotify"));
        }

        if (platform->xev.type == ConfigureNotify) {
            if (this->width != platform->xev.xconfigurerequest.width || this->height != platform->xev.xconfigurerequest.height) {
                Logger::Debug(std::format("Event '{}'", "ResizeRequest"));

                this->width = platform->xev.xconfigurerequest.width;
                this->height = platform->xev.xconfigurerequest.height;

                auto eventData = WindowResizeRequestEvent();
                eventData.Size = { platform->xev.xconfigurerequest.width, platform->xev.xconfigurerequest.height };
                lastKnownWindowSize = eventData.Size;

                OnResizeRequest(eventData);
                OnResizeRequestEvent(eventData);
            }

            //Window Moved.
            if (position.x != platform->xev.xconfigurerequest.x || position.y != platform->xev.xconfigurerequest.y)
                position = { platform->xev.xconfigurerequest.x, platform->xev.xconfigurerequest.y };
        }

    }
    previousKeyboard = currentKeyboard;
    previousMouse.Buttons = currentMouse.Buttons;
}


// Might make the window go off the screen on some window managers.
void RWindow::SetSize(int newWidth, int newHeight) {
    if (!resizable) return;

    this->width = newWidth;
    this->height = newHeight;
    XResizeWindow(platform->display, platform->window, newWidth, newHeight);
    XFlush(platform->display);
    Logger::Info(std::format("Set size for '{}' to {} x {}", this->title, newWidth, newHeight));
}

IPair RWindow::GetAccurateMouseCoordinates() const {

    Window root_return, child_return;
    int root_x_ret, root_y_ret;
    int win_x_ret, win_y_ret;
    uint32_t mask_return;

    // This seems to be relative to the top left corner of the renderable area.
    bool mouseAvailable = XQueryPointer(platform->display, platform->window, &root_return, &child_return, &root_x_ret, &root_y_ret, &win_x_ret, &win_y_ret, &mask_return);

    if (mouseAvailable) {
        // TODO: normalize coordinates from platform->displaySpace to windowSpace
        // TODO: fire mouse movement event
        IPair m_coords = { win_x_ret, win_y_ret };
        return m_coords;
    }
    return {};
}


IPair RWindow::GetSize() const {
    return { this->width, this->height};
}

//IPair RWindow::getLastKnownResize() const
//{
//    return lastKnownWindowSize;
//}

// TODO: implement integer IPair/3 types
IPair RWindow::GetPos() const {
    return position;
}

void RWindow::SetPos(int x, int y) {
    XMoveWindow(platform->display, platform->window, x, y);
    position = {x, y};
}

void RWindow::SetPos(const IPair& pos) {
    SetPos(pos.x, pos.y);
}

void RWindow::Fullscreen() {
    Logger::Info(std::format("Fullscreening '{}'", this->title));
    fullscreen_mode = true;

    XEvent xev;
    Atom wm_state   = XInternAtom(platform->display, "_NET_WM_STATE", true);
    Atom wm_fullscreen = XInternAtom(platform->display, "_NET_WM_STATE_FULLSCREEN", true);

    XChangeProperty(platform->display, platform->window, wm_state, XA_ATOM, 32, PropModeReplace, (unsigned char *)&wm_fullscreen, 1);
    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = platform->window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 1; // _NET_WM_STATE_ADD
    xev.xclient.data.l[1] = fullscreen_mode;
    xev.xclient.data.l[2] = 0;
    XSendEvent(platform->display, DefaultRootWindow(platform->display), False, SubstructureNotifyMask | SubstructureRedirectMask, &xev);
    Logger::Debug(std::format("Fullscreened '{}'", this->title));
}

void RWindow::RestoreFromFullscreen() {
    Logger::Debug(std::format("Restoring '{}' from Fullscreen", this->title));
    fullscreen_mode = false;
    XEvent xev;
    Atom wm_state = XInternAtom(platform->display, "_NET_WM_STATE", False);
    Atom fullscreen = XInternAtom(platform->display, "_NET_WM_STATE_FULLSCREEN", False);
    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = platform->window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 0; // _NET_WM_STATE_REMOVE
    xev.xclient.data.l[1] = fullscreen_mode;
    xev.xclient.data.l[2] = 0;
    XSendEvent(platform->display, DefaultRootWindow(platform->display), False, SubstructureNotifyMask | SubstructureRedirectMask, &xev);
    Logger::Debug(std::format("Restored '{}' from Fullscreen", this->title));
}

void RWindow::SetCursorStyle(CursorStyle style) const {
    auto x11_cursor_resolved_enum = (unsigned int) style.X11Cursor;
    Cursor c = XCreateFontCursor(platform->display, x11_cursor_resolved_enum);
    XDefineCursor(platform->display, platform->window, c);
}

void RWindow::SetTitle(const std::string& title) {
    this->title = title;
    XStoreName(platform->display, platform->window, title.c_str());
}



IPair RWindow::GetPositionOfRenderableArea() const {
    return render_area_position;
}

